/*
 * Copyright 2007 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.hibernate4gwt.core.beanlib.merge;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Map;

import net.sf.beanlib.spi.DetailedBeanPopulatable;
import net.sf.hibernate4gwt.core.IPersistenceUtil;
import net.sf.hibernate4gwt.core.beanlib.BeanlibThreadLocal;
import net.sf.hibernate4gwt.core.store.IPojoStore;
import net.sf.hibernate4gwt.exception.TransientHibernateObjectException;
import net.sf.hibernate4gwt.pojo.base.ILazyPojo;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Populatable for POJO with the lazy information of the Hibernate POJO
 * This populatable is used to fill back an Hibernate POJO from a pure clone
 * POJO (not implementing ILazyPojo).
 * Only the not lazy properties of the Hibernate POJO are filled.
 * To be correct, the clone and merge operations must be managed with the
 * <b>same</b> Hibernate POJO, or a POJO with the <b>same</b> fetching strategy
 * @author bruno.marchesson
 */
public class MergeBeanPopulatable implements DetailedBeanPopulatable 
{
	//----
	// Attributes
	//----
	/**
	 * Log channel
	 */
	private static Log _log = LogFactory.getLog(MergeBeanPopulatable.class);
	
	/**
	 * The associated persistence utils
	 */
	private IPersistenceUtil _persistenceUtil;
	
	/**
	 * The used Pojo store
	 */
	private IPojoStore _pojoStore;
	
	//----
	// Properties
	//----
	/**
	 * @return the persistence Util implementation to use
	 */
	public IPersistenceUtil getPersistenceUtil()
	{
		return _persistenceUtil;
	}

	/**
	 * @param util the persistenceUtil to set
	 */
	public void setPersistenceUtil(IPersistenceUtil util)
	{
		_persistenceUtil = util;
	}
	
	/**
	 * @return the pojo store
	 */
	public IPojoStore getPojoStore()
	{
		return _pojoStore;
	}

	/**
	 * @param store the pojo store to set
	 */
	public void setPojoStore(IPojoStore store)
	{
		_pojoStore = store;
	}

	//-------------------------------------------------------------------------
	//
	// Constructor
	//
	//-------------------------------------------------------------------------
	/**
	 * Constructor
	 */
	public MergeBeanPopulatable(IPersistenceUtil persistenceUtil, IPojoStore pojoStore)
	{
		setPersistenceUtil(persistenceUtil);
		setPojoStore(pojoStore);
	}
	
	//-------------------------------------------------------------------------
	//
	// DetailedBeanPopulatable implementation
	//
	//-------------------------------------------------------------------------	
	/* (non-Javadoc)
	 * @see net.sf.beanlib.spi.DetailedBeanPopulatable#shouldPopulate(java.lang.String, java.lang.Object, java.lang.reflect.Method, java.lang.Object, java.lang.reflect.Method)
	 */
	public boolean shouldPopulate(String propertyName, 
								  Object cloneBean, 
								  Method readerMethod, 
								  Object persistentBean, 
								  Method setterMethod)
	{
		try
		{
		//	Get clone and persistent value
		//
			Object cloneValue = readerMethod.invoke(cloneBean, new Object[] {});
			Object persistentValue = loadPersistentValue(cloneBean, persistentBean, readerMethod);
			
			// Null value : can be a proxy
			if (cloneValue == null)
			{
				if (isLazyProperty(cloneBean, persistentValue, propertyName) == true)
				{
				//	Let the proxy untouched
				//
					return false;
				}
				else
				{
				//	Regular case
				//
					return true;
				}
			}
		
			
			if ((persistentValue instanceof Collection) ||
				(persistentValue instanceof Map))
			{
			//	collection handling
			//
				BeanlibThreadLocal.setDestinationBean(persistentValue);
			}
			else if (_persistenceUtil.isPersistentPojo(persistentValue))
			{
			//	Persistent value : load the associated proxy
			//
				try
				{
					Object loadedProxy = _pojoStore.restore(cloneValue, persistentValue.getClass());
					BeanlibThreadLocal.setDestinationBean(loadedProxy);
				}
				catch(TransientHibernateObjectException e)
				{
				//	Transient value : no proxy
				//
					BeanlibThreadLocal.setDestinationBean(null);
				}
			}
			else
			{
			//	Not a persistent bean : no need for proxy
			//
				BeanlibThreadLocal.setDestinationBean(null);
			}
			
			return true;
		}
		catch (Exception e)
		{
			throw new RuntimeException(e);
		}
	}
	
	
	//-------------------------------------------------------------------------
	//
	// Internal method
	//
	//-------------------------------------------------------------------------
	/**
	 * Load the property from the persistent bean
	 */
	protected Object loadPersistentValue(Object cloneBean, 
										 Object persistentBean,
										 Method readerMethod)
	{
		try
		{
			if (cloneBean.getClass().isAssignableFrom(persistentBean.getClass()))
			{
			//	Invoke the same method
			//
				return readerMethod.invoke(persistentBean, new Object[] {});
			}
			else
			{
			//	Read value with a new method
			//
				Method persistentReaderMethod = persistentBean.getClass().getMethod(readerMethod.getName(), 
																					readerMethod.getParameterTypes());
				return persistentReaderMethod.invoke(persistentBean, new Object[] {});
			}
		}
		catch(Exception ex)
		{
			throw new RuntimeException(ex);
		}
	}
	
	/**
	 * Indicates if the argument property is lazy or not
	 * @param cloneBean
	 * @param persistentBean
	 * @param property
	 * @param readerMethod
	 * @return
	 */
	protected boolean isLazyProperty(Object cloneBean,
									 Object persistentValue,
									 String property)
	{
		if (cloneBean instanceof ILazyPojo)
		{
			return ((ILazyPojo) cloneBean).isLazyProperty(property);
		}
		else
		{
		//	Stateful mode
		//
			return _persistenceUtil.isProxy(persistentValue);
		}
	}
}
